@@ -1,11 +1,12 @@
 package XUL::Gui;
-    use base 'Exporter';
     use strict;
     use warnings;
     use Carp;
-    use Storable qw/dclone/;
-    use List::Util qw/max/;
-    our $VERSION = '0.25';
+    use Exporter 'import';
+    use Storable 'dclone';
+    use List::Util 'max';
+    use MIME::Base64 'encode_base64';
+    our $VERSION = '0.30';
     our $DEBUG = 0;
 
 =head1 NAME
@@ -14,7 +15,7 @@ XUL::Gui - render cross platform gui applications with firefox from perl
 
 =head1 VERSION
 
-version 0.21
+version 0.30
 
 this module is under active development, interfaces may change.
 
@@ -106,28 +107,30 @@ the same way, and can fit into one of four templates
 
 =over 8
 
-=item * C<< HR() is <html:hr /> >>
+=item * C<< HR() is <hr /> >>
 
-=item * C<< B('some bold text') is <html:b>some bold text<html:b/> >>
+=item * C<< B('some bold text') is <b>some bold text<b/> >>
 
 in the special case of a tag with one argument, which is not another tag, that argument is
 added to that tag as a text node. this is mostly useful for HTML tags, but works with XUL as well
 
-=item * C<< Label( value=>'some text', style=>'color: red' ) is <label value="some text" style="color: red;" /> >>
+=item * C<< Label( value=>'some text', style=>'color: red' ) >>
+
+    <label value="some text" style="color: red;" />
 
 =item * C<< Hbox( id=>'mybox', Label('hello'), B('world'), style=>'border: 1px solid black') >>
 
     <hbox id="mybox" style="border: 1px solid black;">
-        <label><textnode value="hello"></label>
-        <html:b>world</html:b>
+        <label>hello</label>
+        <b>world</b>
     </hbox>
 
 unlike XML based XUL, attribute pairs and children can be mixed in any order,
-but should probably be kept at the front for readability
+but attributes should probably be kept at the front for readability
 
 =back
 
-setting the 'id' attribute names the object in the C<%ID> hash.
+setting the 'id' attribute names the object in the global C<%ID> hash.
 otherwise an auto generated name matching C</^xul_\d+$/> is used.
 
     $object = Button( id=>'btn', label=>'OK' );
@@ -150,13 +153,35 @@ javascript event handlers have C<event> and C<this> set for you
         this.label = event.type;
     })
 
-any attribute with a name that doesn't match /^on/ that has a code ref value is added to the object as a method
+any attribute with a name that doesn't match /^on/ that has a code ref value is added
+to the object as a method
+
+Tk's attribute style with a leading dash is supported.
+this is useful for readability when collapsing attribute lists with C<qw//>
+
+    TextBox id=>'txt', width=>75, height=>20, type=>'number', decimalplaces=>4;
+    TextBox qw/-id txt -width 75 -height 20 -type number -decimalplaces 4/;
+
+multiple 'style' attributes are joined with ';' into a single attribute
 
 =head1 EXPORT
 
-    all functions listed here are exported by default, this may change in the future
+    use XUL::Gui;   # is the same as
+    use XUL::Gui qw/:base :util :widget :pragma :xul :html :const :image/;
 
-    all XUL tags: (also exported as Titlecase)
+    the following export tags are available.  :base is required for minimal functionality
+
+    :base       %ID display quit alert function gui XUL
+    :widget     @C %A %M $W widget attribute extends
+    :pragma     buffered now cached noevents delay doevents
+    :const      FLEX FIT FILL SCROLL MIDDLE
+    :image      bitmap bitmap2src serve
+    :util       zip mapn apply trace toggle
+    :internal   tag object genid
+
+    :all    (all exports)
+
+    :xul    (also exported as Titlecase)
         Action ArrowScrollBox Assign BBox Binding Bindings Box Broadcaster BroadcasterSet
         Browser Button Caption CheckBox ColorPicker Column Columns Command CommandSet Conditions
         Content DatePicker Deck Description Dialog DialogHeader DropMarker Editor Grid Grippy
@@ -171,7 +196,7 @@ any attribute with a name that doesn't match /^on/ that has a code ref value is
         ToolBarSpring ToolBox ToolTip Tree TreeCell TreeChildren TreeCol TreeCols TreeItem TreeRow
         TreeSeparator Triple VBox Where Window Wizard WizardPage
 
-    all HTML tags: (also exported as html_lowercase)
+    :html   (also exported as html_lowercase)
         A ABBR ACRONYM ADDRESS APPLET AREA AUDIO B BASE BASEFONT BDO BGSOUND BIG BLINK BLOCKQUOTE
         BODY BR BUTTON CANVAS CAPTION CENTER CITE CODE COL COLGROUP COMMENT DD DEL DFN DIR DIV DL DT
         EM EMBED FIELDSET FONT FORM FRAME FRAMESET H1 H2 H3 H4 H5 H6 HEAD HR HTML I IFRAME ILAYER IMG
@@ -180,11 +205,12 @@ any attribute with a name that doesn't match /^on/ that has a code ref value is
         RT RTC RUBY S SAMP SCRIPT SELECT SMALL SOURCE SPACER SPAN STRIKE STRONG STYLE SUB SUP TABLE
         TBODY TD TEXTAREA TFOOT TH THEAD TITLE TR TT U UL VAR VIDEO WBR XML XMP
 
-    utility:   zip mapn apply trace toggle
-    gui:       display widget attribute extends quit alert dialog function gui XUL
-    constants: FLEX FIT FILL SCROLL MIDDLE
-    pragma:    buffered now cached noevents delay doevents
-    internal:  tag object genid
+    for the OO purists:
+
+        use XUL::Gui ();
+        my $gui = XUL::Gui->new;
+        $gui->display( $gui->Label('hello world') );
+        # use $gui->id('someid') to access the %ID hash;
 
 =cut
 
@@ -212,22 +238,47 @@ any attribute with a name that doesn't match /^on/ that has a code ref value is
         rt rtc ruby s samp script select small source spacer span strike strong style sub sup table
         tbody td textarea tfoot th thead title tr tt u ul var video wbr xml xmp/;
 
-    our @EXPORT = (keys %HTML, @Xul, qw/@C %A %M $W %ID
-        zip mapn apply trace toggle
-        display widget attribute extends quit alert dialog function gui XUL
-        FLEX FIT FILL SCROLL MIDDLE
-        buffered now cached noevents delay doevents
-        tag object genid
-    /);
-
+    our %EXPORT_TAGS = (
+        util     => [qw/zip mapn apply trace toggle/],
+        base     => [qw/%ID display quit alert function gui XUL/],
+        widget   => [qw/@C %A %M $W widget attribute extends/],
+        pragma   => [qw/buffered now cached noevents delay doevents/],
+        xul      => [@Xul],
+        html     => [keys %HTML],
+        const    => [qw/FLEX FIT FILL SCROLL MIDDLE/],
+        image    => [qw/bitmap bitmap2src serve/],
+        internal => [qw/tag object genid/],
+    );
+    our @EXPORT_OK    = map @$_ => values %EXPORT_TAGS;
+    $EXPORT_TAGS{all} = \@EXPORT_OK;
+    our @EXPORT       = map @{ $EXPORT_TAGS{$_} } =>
+                            qw/util base widget pragma xul html const image/;
     our %defaults = (
         window      => ['xmlns:html' => 'http://www.w3.org/1999/xhtml',
                          xmlns       => 'http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul' ],
         scrollbox   => [ scrollTo    => sub {my ($self, $x, $y) = @_; gui("scrollTo(ID.$$self{ID}, $x, $y);")} ],
+        textbox     => [  value      => sub :lvalue {tie my $ret, 'XUL::Gui::Scalar', shift, '_value'; $ret},
+                         _value      => sub :lvalue {tie my $ret, 'XUL::Gui::Scalar', shift,  'value'; $ret} ],
     );
-    our $server = XUL::Gui::Server->init;
-    our (%ID, %_ID);
-    our %dialogs;
+    our $server ||= XUL::Gui::Server->init;
+    our (%ID, %_ID, %dialogs);
+
+{package
+    XUL::Gui::OO;
+    use strict;
+    use warnings;
+    our %lookup = map {lc, $_} @{ $EXPORT_TAGS{xul} };
+    sub AUTOLOAD {
+        shift;
+        no strict 'refs';
+        my ($sub) = our $AUTOLOAD =~ /([^:]+)$/;
+        goto &{'XUL::Gui::' . ($lookup{lc $sub} || $sub)};
+    }
+    sub DESTROY {}
+    sub id :lvalue {$XUL::Gui::ID{$_[1]}}
+    sub ID :lvalue {$XUL::Gui::ID{$_[1]}}
+}
+    sub new {bless {} => 'XUL::Gui::OO'}
 
 =head1 FUNCTIONS
 
@@ -244,12 +295,16 @@ map over n elements at a time in C<@_> with C<$_ == $_[0]>
 
 =cut
     sub mapn (&$@) {
-        my ($c, $by, @r) = splice @_, 0, 2;
+        my ($sub, $n, @ret) = splice @_, 0, 2;
+        croak '$_[1] must be >= 1' unless $n >= 1;
+        my $ret = defined wantarray;
         while (@_) {
             local *_ = \$_[0];
-            push @r, $c->( splice @_, 0, $by )
+            if ($ret) {push @ret =>
+                  $sub->(splice @_, 0, $n)}
+            else {$sub->(splice @_, 0, $n)}
         }
-        @r
+        @ret
     }
 
 =item C<zip LIST of ARRAYREF>
@@ -259,13 +314,13 @@ map over n elements at a time in C<@_> with C<$_ == $_[0]>
 =cut
     sub zip {
         map {my $i = $_;
-            map {$$_[$i]} @_
+            map $$_[$i] => @_
         } 0 .. max map $#$_ => @_
     }
 
 =item C<apply {CODE} LIST>
 
-apply a function to a list and return that list
+apply a function to a copy of LIST and return the copy
 
     print join ", " => apply {s/$/ one/} "this", "and that";
     > this one, and that one
@@ -290,17 +345,69 @@ alternate a variable between two states
         $_[0] = $_[ 1 + ($_[0] eq $_[1] or $_[0] ne $_[2]) ]
     }
 
+=item C<bitmap WIDTH HEIGHT OCTETS>
+
+returns a binary .bmp bitmap image. OCTETS is a list of BGR values
+
+    bitmap 2, 2, qw(255 0 0 255 0 0 255 0 0 255 0 0); # 2px blue square
+
+=cut
+
+    sub bitmap {
+        my ($width, $height, @bitmap) = splice @_, 0, 2;
+        my $pad = $width % 4;
+        $pad = 4 - $pad if $pad;
+        for (1 .. $height) {
+            unshift @bitmap, splice( @_, 0, $width*3 ), (0) x $pad
+        }
+        pack 'n V n n (N)2 (V)2 n n N V (N)4 (C)*' =>
+            0x42_4D,         # "BM"
+            (54 + @bitmap),  # file size
+            0x00_00,         # not used
+            0x00_00,         # not used
+            0x36_00_00_00,   # offset of bitmap data
+            0x28_00_00_00,   # remaining bytes in header
+            $width,
+            $height,
+            0x01_00,         # color planes (always 1)
+            0x18_00,         # bits/pixel (24)
+            0x00_00_00_00,   # no compression
+            scalar(@bitmap), # size of raw BMP data (after header)
+            0x13_0B_00_00,   # horizontal resolution (2,835 pixels/meter)
+            0x13_0B_00_00,   # vertical resolution (2,835 pixels/meter)
+            0x00_00_00_00,   # not used
+            0x00_00_00_00,   # not used
+            @bitmap
+    }
+
+
+=item C<bitmap2src WIDTH HEIGHT OCTETS>
+
+returns a packaged bitmap image that can be directly assigned to an image
+tag's src attribute. arguments are the same as C<bitmap()>
+
+    $ID{myimage}->src = bitmap2src 320, 180, @image_data;
+
+=cut
+
+    sub bitmap2src {
+        'data:image/bitmap;base64,' . encode_base64 &bitmap
+    }
+
     {my $id = 0;
         sub genid () {'xul_' . $id++}
     }
 
-    sub isa {UNIVERSAL::isa @_ > 1 ? shift : $_, @_}
+    sub isa {UNIVERSAL::isa( (@_, $_)[1, 0] )}
 
     sub parse {
         my (@C, %A, %M);
         while (local $_ = shift) {
-            if (isa 'XUL::Gui::Object') {push @C, $_}
-            elsif (/^on/ or ref $_[0] ne 'CODE') {
+            if (isa 'XUL::Gui::Object') {push @C, $_; next}
+            grep {not defined and $_ = '???'} $_, $_[0]
+                and croak "parse failure: [ $_ => $_[0] ] @_[1..$#_],";
+            s/^-//;
+            if (/^on/ or ref $_[0] ne 'CODE') {
                 /^style$/ and $A{$_} .= (shift).';'
                           or  $A{$_}  =  shift}
             else             {$M{$_}  =  shift}
@@ -327,7 +434,7 @@ alternate a variable between two states
 =cut
 
 sub FLEX   {flex  => 1,                   @_}
-sub FILL   {align => 'stretch', FLEX      @_}
+sub FILL   {align => 'stretch', FLEX,     @_}
 sub FIT    {sizeToContent => 1,           @_}
 sub SCROLL {style => 'overflow: auto',    @_}
 sub MIDDLE {qw/align center pack center/, @_}
@@ -350,8 +457,8 @@ takes a list of gui objects, and several optional parameters:
     port      (8888)     first port to try starting the server on, port++ after that
     delay  milliseconds  delays each gui update cycle
 
-if C<$_[0]> is a C<Window>, that window is created, otherwise a default one is added.
-gui objects from C<@_> are then added to the window.
+if the first object is a C<Window>, that window is created, otherwise a default
+one is added. the remaining objects are then added to the window.
 
 C<display> will not return until the the gui quits
 
@@ -360,7 +467,7 @@ see SYNOPSYS and XUL::Gui::Manual for more details
 =cut
     sub display {$server->start( &parse )}
 
-    sub dialog {carp 'dialog not implemented yet'}
+    sub dialog {croak 'dialog not implemented yet'}
 
 =item C<quit>
 
@@ -368,10 +475,21 @@ shuts down the server (causes a call to C<display> to return at the end of the c
 
 =cut
     sub quit {
-        gui('quit();');
+        gui('quit();') unless @_;
         $$server{run} = 0;
     }
 
+=item C<serve PATH MIMETYPE DATA>
+
+add a virtual file to the server
+
+    serve '/myfile.jpg', 'text/jpeg', $jpegdata;
+
+the paths C<qw( / /client.js /event /ping /exit /perl )> are reserved
+
+=cut
+    sub serve {$server->serve(@_)}
+
 =item C<object TAGNAME LIST>
 
 creates a gui proxy object, allows run time addition of custom tags
@@ -394,7 +512,7 @@ creates a gui proxy object, allows run time addition of custom tags
 
 =item C<tag NAME>
 
-returns a CODEREF that generates proxy objects, allows for user defined tag functions
+returns a code ref that generates proxy objects, allows for user defined tag functions
 
     *mylabel = tag 'label';
 
@@ -402,10 +520,12 @@ returns a CODEREF that generates proxy objects, allows for user defined tag func
 
 =cut
     sub tag {
-        my @arg = @_;
+        my @args = @_;
         sub {
-            unshift @_, 'TEXT' if @_ == 1 and not isa $_[0] => 'XUL::Gui::Object';
-            object @arg, @_
+            object @args,
+                (@_ == 1 and not isa 'XUL::Gui::Object' => $_[0])
+                    ? 'TEXT' : (),
+                @_
         }
     }
 
@@ -468,7 +588,7 @@ group tags together into common patterns, with methods and inheritance
 
             for my $i (keys %ID) { # refactor to reduce copying
                 $_ID{$wid}{$i} = $_ID{ my $gid = genid } = $ID{$i};
-                next unless isa $ID{$i} => 'XUL::Gui::Object';
+                next unless isa 'XUL::Gui::Object' => $ID{$i};
                 $ID{$i}{N}    = $ID{$i}{A}{id};
                 $ID{$i}{ID}   = $ID{$i}{A}{id} = $gid;
                 $ID{$i}{W}    = $_ID{$wid};
@@ -502,13 +622,23 @@ includes an attribute name if it exists, only works inside of widgets.
 NAME is split on whitespace
 
     attribute 'label type' # is syntactic sugar for
-    map {$_ => $A{$_} grep {exists $A{$_}} qw/label type/
+    map {$_ => $A{$_}} grep {exists $A{$_}} qw/label type/
+
+    attribute '+' # same as %A
+    attribute '*' # any untouched attributes (by attribute)
 
 =cut
     sub attribute ($) {
         croak 'attribute only works inside widgets' unless defined %_ID;
-        my ($key, $A) = (shift, (caller).'::A');
-        map {$_ => $$A{$_}} grep {exists $$A{$_}} split /\s+/ => $key
+        my ($A, $W, @ret) = ((caller).'::A', (caller).'::W');
+        my @keys = grep {
+            $_ eq '*' ? do {push @ret, %$A; 0} :
+            $_ eq '+' ? do {push @ret, map {$_ => $$A{$_}}
+                grep {not $$W{SEEN}{$_}} keys %$A; 0
+            } : 1
+        } split /\s+/ => shift;
+        $$W{SEEN}{$_}++ for @keys;
+        map {$_ => $$A{$_}} grep {exists $$A{$_}} @keys
     }
 
 
@@ -520,7 +650,7 @@ this function is provided to facilitate drag and drop of XML based XUL from tuto
 the perl functional syntax for tags should be used in all other cases
 
 =cut
-    {my %xul; @xul{map {lc} @Xul} = @Xul;
+    {my %xul; @xul{map lc, @Xul} = @Xul;
     sub XUL {
         for ("@_") {
             s {<(\w+)(.+?)}       "$xul{lc $1}($2"g;
@@ -541,7 +671,7 @@ open an alert message box
 =cut
     sub alert {
         gui( "alert('\Q@_\E');" );
-        @_
+        wantarray ? @_ : pop
     }
 
 =item C<trace LIST>
@@ -565,14 +695,14 @@ carps LIST with object details, and then returns LIST unchanged
         no warnings;
         our %space;
         local *space = \%{"$_[0]\::"};
-        for (keys %space) { eval {
-            *{$space{$_}}{CODE} == ($$self{T} || $$self{W}{T})
+        for (keys %space) {
+            eval {*{$space{$_}}{CODE}} == ($$self{T} || $$self{W}{T})
                 and return $cache{$self} = $_ .
                     ($$self{T}
                         ? '{'
-                        : "{".($$self{W}{A}{id} || $$self{W}{ID}).'}->{'
+                        : '{'.($$self{W}{A}{id} || $$self{W}{ID}).'}->{'
                     ).($$self{N} || $$self{ID}).'}'
-        }}
+        }
         $$self{ID}
     }}
 
@@ -582,14 +712,13 @@ carps LIST with object details, and then returns LIST unchanged
         my $self = $_[0];
         my $caller = caller 1;
         our $AUTOLOAD =~ /([^:]+)$/;
-        lookup($self, $caller) . "->$1(".
+        lookup($self, $caller) . "->$1(" .
             (join ',' => map {
                 (isa 'XUL::Gui::Object') ? lookup($_, $caller) : $_
-            } @_[1..$#_]). ")\n";
+            } @_[1..$#_]) . ")\n";
     }
 
 
-
 =item C<function JAVASCRIPT>
 
 create a javascript function, useful for functions that need to be very fast, such as rollovers
@@ -599,24 +728,24 @@ create a javascript function, useful for functions that need to be very fast, su
         alert('hello from javascript');
     })
 
+    to access widget siblings by id, wrap the id with C< W{...} >
+
 =cut
     sub function ($) {
-        (my $js = shift) =~ s[\$?W{\s*(\w+)\s*}]
-                             [ID.\$_->{W}{$1}{ID}]g;
+        my $js = shift;
         bless [ sub {
-            my $id = shift;
+            my $self = shift;
             my $func = 'ID.' . genid;
-            delay( sub{
-                local *_ = \$ID{$id};
-                gui( "$func = function(event){ (function(){ ". eval(qq/"$js"/) ." }).call( ID.$id ) }" );
+            delay( sub {
+                $js =~ s[\$?W{\s*(\w+)\s*}] [ID.$$self{W}{$1}{ID}]g;
+                gui( "$func = function (event) {
+                    (function(){ $js }).call( ID.$$self{ID} )
+                }" )
             });
-            "$func(event)";
+            "$func(event)"
         } ] => 'XUL::Gui::Function'
     }
 
-
-
-
 =item C<gui JAVASCRIPT>
 
 executes JAVASCRIPT
@@ -626,15 +755,15 @@ executes JAVASCRIPT
 =cut
 
     {my ($buffered, @buffer, $cached, %cache, $now);
-        sub gui : lvalue {
+        sub gui :lvalue {
             push @_, "\n";
             unless ($now) {
                 push @buffer, @_ and return if $buffered;
                 return $cache{$_[0]} if exists $cache{$_[0]}
             }
             $server->write('text/plain', "@_");
-            my $res = $server->read->{CONTENT};
-            croak "Invalid Response: $res" unless $res =~ /^(...) (.*)/;
+            my $res = $server->safe_read->{CONTENT};
+            croak "invalid response: $res" unless $res =~ /^(...) (.*)/;
 
             $res = $1 eq 'OBJ' ?
                 ($ID{$2} || object undef, id=>$2) : $2;
@@ -703,7 +832,7 @@ run after the gui objects are rendered
     sub delay (&@) {
         my $code = shift;
         my @args = @_;
-        push @{$$server{queue}}, sub{ $code->( @args ) };
+        push @{$$server{queue}}, sub{ $code->(@args) };
         return;
     }
 
@@ -728,7 +857,7 @@ force a gui update before an event handler finishes
 =cut
     sub doevents () {
         $server->write('text/plain', 'NOOP');
-        $server->read;
+        $server->read; # should be safe_read
         return;
     }
 
@@ -736,17 +865,18 @@ force a gui update before an event handler finishes
 
     # access attributes and properties
 
-        $object->value = 5;     # sets the value in the gui
-        print $object->value;   # gets the value from the gui
+        $object->value = 5;                     # sets the value in the gui
+        print $object->value;                   # gets the value from the gui
 
     # the attribute is set if it exists, otherwise the property is set
 
-        $object->_value = 7;    # sets the property directly
+        $object->_value = 7;                    # sets the property directly
 
-    # function calls
+    # method calls
 
         $object->focus;                         # void context
-        $object->appendChild( H2('title') );    # or any arguments are always function calls
+        $object->appendChild( H2('title') );    # or any arguments are always method calls
+        print $object->someAccessorMethod_;     # append _ to force interpretation as a JS method call
 
 in addition to mirroring all of an object's existing javascript methods / attributes / and properties
 to perl (with identical spelling / capitalization), several default methods have been added to all objects
@@ -767,7 +897,7 @@ package
         }
     };
 
-    sub AUTOLOAD : lvalue {
+    sub AUTOLOAD :lvalue {
         my $self = $_[0];
         return unless (my $AL) = our $AUTOLOAD =~ /([^:]+)$/;
 
@@ -779,22 +909,22 @@ package
             if ($XUL::Gui::DEBUG) {
                 my $caller = caller;
                 print XUL::Gui::lookup($self, $caller) . "->$1(" .
-                    (join ',' => map {(XUL::Gui::isa XUL::Gui::Object) ? XUL::Gui::lookup($_, $caller) : $_} @_[1..$#_]). ")\n"
+                    (join ',' => map {(XUL::Gui::isa XUL::Gui::Object)
+                        ? XUL::Gui::lookup($_, $caller) : $_} @_[1..$#_]). ")\n"
             }
             goto &$method
         }
         shift;
-        if (@_ or not defined wantarray) { # js method call
+        if ($AL =~ s/_$// or @_ or not defined wantarray) { # js method call
             my ($js, $arg) = ('') x 2;
             $_->($self), return for $$self{uc $AL} or ();
-            shift if @_ and $_[0] eq '_';
             $arg = join ',', map {
-                XUL::Gui::isa XUL::Gui::Object and do
-                    {if ($_->{DIRTY}){
+                XUL::Gui::isa XUL::Gui::Object and do {
+                    if ($_->{DIRTY}) {
                         ($$_{W} ? $$_{W}{W} : $$_{W}) ||= $$self{W};
                         $$_{P} ||= $self;
-                        $js .= $_->toJS;}
-                    "ID.$_->{ID}"}   or   "'\Q$_\E'"
+                        $js .= $_->toJS;
+                    } "ID.$_->{ID}" }  or  "'\Q$_\E'"
                 } @_;
             return XUL::Gui::gui($js . "ID.$self->{ID}.$AL($arg);")
         }
@@ -829,8 +959,9 @@ package
             push @xul, ">\n";
             push @xul, "\t" x ($tab+1),  $_->toXUL($tab+1) for @{$$self{C}};
             push @xul, "\t" x $tab, "</$self->{TAG}>\n";
+        } else {
+            push @xul, "/>\n"
         }
-        else {push @xul, "/>\n"}
         join '' => @xul
     }
 
@@ -851,16 +982,16 @@ package
         for (keys %{$$self{A}}) {
             my $val = quotemeta(
                 ref $$self{A}{$_} eq 'XUL::Gui::Function'
-                    ? $$self{A}{$_}[0]( $$self{ID} )
+                    ? $$self{A}{$_}[0]( $self )
                     : $$self{A}{$_}
             );
             if (/^TEXT$/) {
                 push @js, qq{$id.appendChild( document.createTextNode('$val') );};
                 next
             }
-            push @js, /^_(.+)/ ?
-                qq{$id\['$1'] = '$val';} :
-                qq{$id.setAttribute('\L$_\E','$val');};
+            push @js, /^_(.+)/
+                        ? qq{$id\['$1'] = '$val';}
+                        : qq{$id.setAttribute('\L$_\E','$val');};
         }
         for (@{$$self{C}}) {
             $$_{P} = $self;
@@ -912,7 +1043,6 @@ inserts CHILD at INDEX (defaults to 0) in the parent's child list
     sub prependChild {
         my ($self, $child, $count, $first) = @_;
         if ($$self{TAG} eq 'tabs') {
-            print "tabs: ";
             $first = $self->getItemAtIndex( $count || 0 )
         } else {
             $first = $self->firstChild;
@@ -921,7 +1051,6 @@ inserts CHILD at INDEX (defaults to 0) in the parent's child list
                 $first = $first->nextSibling;
             }
         }
-        print "$$first{TAG}\n\n" if $first ne 'null';
         $first eq 'null'
             ? $self->appendChild( $child )
             : $self->insertBefore( $child, $first );
@@ -962,18 +1091,14 @@ removes all items, then appends LIST
     }
 
 
-
 package
     XUL::Gui::Scalar;
-    use base 'XUL::Gui::Object';
+    our @ISA = 'XUL::Gui::Object';
     use warnings;
     use strict;
     use Carp;
 
-    sub TIESCALAR {
-        my $class = shift;
-        bless [@_] => $class;
-    }
+    sub TIESCALAR {bless [ @_[1..$#_] ] => $_[0]}
 
     sub FETCH {
         my ($self, $AL) = @{+shift};
@@ -986,9 +1111,14 @@ package
     sub STORE {
         my ($self, $AL, $new) = (@{+shift}, @_);
         if ($AL =~ /^on/) {
-            not defined $new or ref $new eq 'CODE'
-                or croak "assignment to event handler must be CODE ref or undef";
-            $new = $new ? do {$$self{uc $AL} = $new; 'EVT(event)'} : '';
+            if (ref $new eq 'XUL::Gui::Function') {
+                XUL::Gui::gui $$new[0];
+                $new = $$new[1]
+            } else {
+                not defined $new or ref $new eq 'CODE'
+                    or croak "assignment to event handler must be CODE ref or undef";
+                $new = $new ? do {$$self{uc $AL} = $new; 'EVT(event)'} : '';
+            }
         }
         XUL::Gui::gui $AL =~ /^_(.+)/
             ? "1;ID.$self->{ID}\['$1'] = '\Q$new\E';"
@@ -1023,8 +1153,10 @@ package
 
         @$self{keys %params} = values %params;
 
+        $$self{caller} = caller 1;
+
         message "version $VERSION" if
-            local $DEBUG = $params{debug} || $DEBUG;;
+            local $DEBUG = $params{debug} || $DEBUG;
 
         my $port = $params{port} || $port++;
         $port++ until $$self{server} = IO::Socket::INET->new(
@@ -1039,8 +1171,9 @@ package
         $| = 1;
         message "server started: http://localhost:$port";
 
-        my ($root, %dispatch);
-        %dispatch = (
+        my $root;
+        $$self{dispatch} = {
+            exists $$self{dispatch} ? %{$$self{dispatch}} : (),
             '/' => sub {
                 my $res = qq{<?xml version="1.0"?>\n<?xml-stylesheet }.
                           qq{href="chrome://global/skin" type="text/css"?>\n};
@@ -1049,7 +1182,8 @@ package
                             ? shift @content
                             : XUL::Gui::Window();
 
-                $root->{A}{onunload} = sub {$$self{run} = 0};
+                $$self{onunload} ||= $$root{A}{onunload};
+                $$root{A}{onunload} = q{ send('exit','') };
                 unshift @content, @{ $root->{C} };
                 $root->{C} = [ XUL::Gui::Script(src=>'/client.js') ];
                 $self->write('application/vnd.mozilla.xul+xml', $res . $root->toXUL);
@@ -1068,7 +1202,7 @@ package
                 $evt = "ON\U$evt";
                 $id  = $root->{ID} if $id eq 'undefined'; # why doesn't onunload window send id?
                 $id  = $ID{$id}{P}{ID} while $ID{$id}{P} and not $ID{$id}{$evt}; # more xul bugs
-                if (ref $ID{$id}{$evt} eq 'CODE'){
+                if (ref $ID{$id}{$evt} eq 'CODE') {
                     $ID{$id}{$evt} -> ( $ID{$id}, XUL::Gui::object(undef, id=>$obj) )
                 } else {message "no event handler found: $req->{CONTENT}"}
                 $self->write('text/plain', 'NOOP');
@@ -1082,8 +1216,15 @@ package
             '/favicon.ico' => sub {
                 $self->write('text/plain', '');
             },
-            '/exit' => sub {$$self{run} = 0}
-        );
+            '/exit' => sub {
+                $$self{run} = 0;
+                no warnings;
+                local @CARP_NOT = qw/XUL::Gui XUL::Gui::Server XUL::Gui::Object/;
+                local *XUL::Gui::gui = sub {croak 'gui not available'};
+                use warnings;
+                ref eq 'CODE' and $_->() for $$self{onunload};
+            }
+        };
         unless ($params{nolaunch}) {
             if ($^O =~ /darwin/) {
                 system qq[osascript -e 'tell application "Firefox" to OpenURL "http://localhost:$port"']
@@ -1093,44 +1234,59 @@ package
                     ($^O =~ /MSWin/)
                         ? (map {chomp; "$_\\"} split ',' => `echo \%ProgramFiles\%,\%ProgramFiles(x86)\%`)
                         : split /[:;]/ => $ENV{PATH}
-                ){
-                    find sub{ push @firefox, [$_, $File::Find::name] if /^firefox(?:-bin|\.exe)?$/ and -f } => $_
+                ) {
+                    find sub {push @firefox, [$_, $File::Find::name] if /^firefox(?:-bin|\.exe)?$/ and -f} => $_
                         for grep {/mozilla|firefox/i} map {"$dir/$_"} grep -d, (-d $dir and chdir $dir) ? <*> : ()
                 }
-
                 if (@firefox = sort {length $$a[0] < length $$b[0]} @firefox) {
-                    my $insert  = $params{nochrome} ? '' : ' -chrome ';
                     message 'launching firefox';
                     unless ($$self{pid} = fork) {
                         $firefox[0][1] =~ tr./.\\. if $^O =~ /MSWin/;
-                        exec qq{"$firefox[0][1]" $insert "http://localhost:$port" }
-                             .(q{1>&2 2>/dev/null} x ($^O !~ /MSWin/))
+                        exec qq{"$firefox[0][1]" }
+                            . ($params{nochrome} ? '' : '-chrome ')
+                            . ($params{trusted}
+                                ? qq{"chrome://xulgui/content/xulgui.xul?$port"}
+                                : qq{"http://localhost:$port"}
+                            ).(q{ 1>&2 2>/dev/null} x ($^O !~ /MSWin/))
                     }
                 }
                 else {local $silent; message 'firefox not found: start manually'}
             }
         }
-         $$self{run} = 1;
-         while ($$self{client} = $$self{server}->accept) {
-                $$self{client}->autoflush(1);
-                message 'client connected';
-                while (local $req = $self->read) {
-                    if ($dispatch{$$req{URL}}) {
-                        $dispatch{$$req{URL}}->()
-                    } elsif (open FILE, '.'.$$req{URL}) {
-                        $self->write('text/plain', <FILE>)
-                    } else {
-                        message "file: $$req{URL} not found";
-                        $self->write('text/plain', '')
-                    }
-                    last unless $$self{run}
-                }
-                close $$self{client};
+        $$self{run} = 1;
+        $$self{cycle} = sub {
+            if ($$self{dispatch}{$$req{URL}}) {
+                $$self{dispatch}{$$req{URL}}->()
+            } elsif (open FILE, '.'.$$req{URL}) {
+                $self->write('text/plain', do{local $/; <FILE>})
+            } else {
+                message "file: $$req{URL} not found";
+                $self->write('text/plain', '')
+            }
+        };
+        while ($$self{client} = $$self{server}->accept) {
+            $$self{client}->autoflush(1);
+            message 'client connected';
+            while (local $req = $self->read) {
+                $$self{cycle}->();
                 last unless $$self{run}
             }
+            close $$self{client};
+            last unless $$self{run}
+        }
         $self->stop('server stopped')
     }
 
+    sub safe_read {
+        my $self = shift;
+        while (local $req = $self->read) {
+            message "safe read $$req{URL}" if $DEBUG > 2;
+            return $req if $$req{URL} eq '/res';
+            $$self{cycle}->();
+            return unless $$self{run}
+        }
+    }
+
     sub read {
         my ($self, $client, %req) = ($_[0], $_[0]{client});
         local ($/, $_) = "\015\012";
@@ -1146,13 +1302,13 @@ package
                 {usleep 1000*$$self{delay} unless m|/ping|}
             unless (defined) {
                 $self->write( 'text/javascript', 'alert("error: broken message received")' );
-                message 'broken message received';
-                return $self->read
+                die 'broken message received';
             }
-            ($_, $req{CONTENT}) = ($1, "OBJ $2")  if  /(.+?)\?(.+)/;
+            ($_, $req{CONTENT}) = ($1, "OBJ $2")  if  /(.+?)\?(.*)/;
             message "read $_ $req{CONTENT}"       if  $DEBUG > 2 && !m|/ping|;
             if (m|/perl|) {
-                $self->write( 'text/plain', 'RETURN '. eval $req{CONTENT} );
+                $self->write( 'text/plain', 'RETURN '.
+                    (eval("no strict;package $$self{caller}; $req{CONTENT}") || '' ));
                 %req = %{ $self->read }
             }
         }
@@ -1160,17 +1316,16 @@ package
     }
 
     sub write {
-        my $client = shift->{client};
-        message "write @_" if $DEBUG > 2;
-        print $client join "\015\012" =>
+        message "write @_[1,2]" if $DEBUG > 2;
+        print {$_[0]{client}} join "\015\012" =>
             'HTTP/1.1 200 OK',
             'Expires: -1',
             'Keep-Alive: 300',
             'Connection: Keep-Alive',
-            'Content-type: ' . shift,
-            'Content-length: ' . length,
+            'Content-type: ' . $_[1],
+            'Content-length: ' . length $_[2],
             '',
-            $_ for "@_[1..$#_]"
+            $_[2]
     }
 
     sub stop {
@@ -1179,6 +1334,16 @@ package
         kill 1, -($^O =~ /MSWin/ ? shift->{pid} : $$);
     }
 
+    sub serve {
+        my ($self, $path, $type, $data) = @_;
+        $path =~ m[^/(?:client.js|event|ping|exit)?$]
+            and croak 'reserved path';
+        $$self{dispatch}{$path} = sub {
+            $self->write($type, $data)
+        };
+        $path
+    }
+
     $client_js = <<'END';
 
 Object.prototype.__defineGetter__('__sid__', function () {
@@ -1255,36 +1420,39 @@ var noEvents    = {};
 var cacheEvents = true;
 var ping        = 250;
 var interval    = setInterval( pinger, ping );
-var server      = 'http://localhost:' + port + '/';
+var host        = 'http://localhost:' + port + '/';
 var queue       = [];
 var mutex       = false;
-var perl        = new XMLHttpRequest();
+var server      = new XMLHttpRequest();
 
 function pinger () {
     if (mutex || !cacheEvents) return;
     mutex = true;
-    send('ping', null)
+    send('ping', null);
     mutex = false;
 }
 
 var retre = new RegExp(/^RETURN (.*)/);
 function send (to, val) {
-    var url = server + to;
+    var url = host + to;
+    var resurl = host + 'res';
     var type;
     while (1) {
-        perl.open( 'POST', url, false );
-        perl.send( val );
-        if (perl.responseText != 'NOOP') {
-            var ret = perl.responseText.match(retre);
+        server.open( 'POST', url, false );
+        server.send( val );
+        if (server.responseText != 'NOOP') {
+            var ret = server.responseText.match(retre);
             if (ret) return ret[1];
-            try {val = eval( perl.responseText )}
-            catch(e) {alert( [ e.name, perl.responseText, e.message ].join("\n\n") )}
+            try {val = eval( server.responseText )}
+            catch(e) {alert( [ e.name, server.responseText, e.message ].join("\n\n") )}
 
             type = typeof val;
 
             if (val == 0 && type != 'string') {val = '0'}
             else if (val == null) {val = 'null'}
 
+            url = resurl;
+
             if (type == 'object') {
                 if (val.hasAttribute && val.hasAttribute('id'))
                     {val = 'OBJ ' + val.getAttribute('id')}
@@ -1299,13 +1467,44 @@ function send (to, val) {
     }
     return val;
 }
+/*
+function send (to, val) {
+    perl.open( 'POST', server + to );
+    perl.send( val );
+}
 
+perl.onreadystatechange = function () {
+    if (perl.readyState != 4) return;
+    var val;
+    if (perl.responseText != 'NOOP') {
+        var ret = perl.responseText.match(retre);
+        if (ret) return ret[1];
+        try {val = eval( perl.responseText )}
+        catch(e) {alert( [ e.name, perl.responseText, e.message ].join("\n\n") )}
 
-var xul_id = function (obj) {
-    while (obj.parentNode && obj.parentNode != obj) {
-        if(obj.id) return obj.id
-        obj = obj.parentNode
+        var type = typeof val;
+
+        if (val == 0 && type != 'string') {val = '0'}
+        else if (val == null) {val = 'null'}
+
+        if (type == 'object') {
+            if (val.hasAttribute && val.hasAttribute('id'))
+                {val = 'OBJ ' + val.getAttribute('id')}
+            else {
+                ID[ 'xul_js_' + id ] = val;
+                val = 'OBJ xul_js_' + id++;
+            }
+        }
+        else {val = 'RES ' + val}
+        send('res', val)
     }
+}
+*/
+var xul_id = function (obj) {
+    do {
+        if (obj.id) return obj.id
+        obj = obj.parentNode
+    } while (obj.parentNode && obj.parentNode != obj);
 }.cache(true);
 
 function EVT (event) {
@@ -1318,8 +1517,13 @@ function EVT (event) {
     mutex = true;
     do {
         if (event) {
-            ID['xul_js_' + id] = event;
-            send('event', "EVT " + xul_id(event.target) + " " + event.type + " " + ('xul_js_' + id++));
+            if (event.type == 'perl') {
+                send('perl', event.code)
+            } else {
+                ID['xul_js_' + id] = event;
+                send('event', 'EVT ' + xul_id(event.target) +
+                     ' ' + event.type + ' ' + ('xul_js_' + id++));
+            }
         } else {
             send('ping', null)
         }
@@ -1368,9 +1572,11 @@ function scrollTo (element, x, y) {
     ).scrollTo(x, y);
 }
 
-function Perl (code) {
-    return send('perl', code)
+function perl (code) {
+    return mutex ? send('perl', code)
+                 : EVT({ type: 'perl', code: code })
 }
+var Perl = perl;
 
 function dialog (target, code) {
     return target.install(code)
@@ -1402,6 +1608,9 @@ todo list.
 
 the code that attempts to find firefox may not work in all cases, patches welcome
 
+for the TextBox object, the behaviors of the "value" and "_value" methods are reversed.
+it works better that way and is more consistent with the behavior of other tags.
+
 =head1 AUTHOR
 
 Eric Strom, C<< <ejstrom at gmail.com> >>
@@ -1412,7 +1621,7 @@ please report any bugs or feature requests to C<bug-xul-gui at rt.cpan.org>, or
 the web interface at L<http://rt.cpan.org/NoAuth/ReportBug.html?Queue=XUL-Gui>.
 I will be notified, and then you'll automatically be notified of progress on your bug as I make changes.
 
-=head1 ACKNOWLEDGEMENTS
+=head1 ACKNOWLEDGMENTS
 
 the mozilla development team
 
